package ru.dbotthepony.kstarbound.client.gl.vertex

import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import org.lwjgl.opengl.GL11.GL_UNSIGNED_INT
import org.lwjgl.opengl.GL11.GL_UNSIGNED_SHORT
import ru.dbotthepony.kstarbound.client.gl.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.GLVertexBufferObject
import ru.dbotthepony.kstarbound.util.readLEShort
import ru.dbotthepony.kstarbound.util.writeLEInt
import java.nio.ByteBuffer
import java.nio.ByteOrder

/**
 * Vertex Builder который хранит все данные на куче, при [upload] создаётся [ByteBuffer] вне кучи,
 * в него записываются данные из кучи, данные загружаются на видеокарту.
 *
 * Полезен в случаях, когда размер данных для загрузки заранее не известен.
 */
class HeapVertexBuilder(
	attributes: GLAttributeList,
	type: VertexType,
) : AbstractVertexBuilder<HeapVertexBuilder>(attributes, type) {
	override val vertexMemory = FastByteArrayOutputStream()
	override val elementMemory = FastByteArrayOutputStream()

	override var elementIndexType: Int = GL_UNSIGNED_SHORT
		private set

	override fun ensureIndexCapacity() {
		if (elementIndexType == GL_UNSIGNED_SHORT && elementMemory.length / 2 + type.indicies.size >= 30000) {
			val backing = elementMemory.array
			val copy = FastByteArrayInputStream(ByteArray(elementMemory.length) { backing[it] })

			elementIndexType = GL_UNSIGNED_INT
			val elementMemory = elementMemory
			elementMemory.reset()

			for (i in 0 until (copy.length / 2)) {
				elementMemory.writeLEInt(copy.readLEShort())
			}
		}
	}

	override fun resetMemory() {
		vertexMemory.reset()
		elementMemory.reset()
	}

	override fun doUpload(vbo: GLVertexBufferObject, ebo: GLVertexBufferObject, drawType: Int) {
		val vboMemory = ByteBuffer.allocateDirect(vertexMemory.length)
		vboMemory.put(vertexMemory.array, 0, vertexMemory.length)

		val eboMemory = ByteBuffer.allocateDirect(elementMemory.length)
		eboMemory.put(elementMemory.array, 0, elementMemory.length)

		vboMemory.position(0)
		eboMemory.position(0)

		vbo.bufferData(vboMemory, drawType)
		ebo.bufferData(eboMemory, drawType)
	}
}
