package ru.dbotthepony.kstarbound.world.phys

import ru.dbotthepony.kstarbound.world.CHUNK_SHIFT
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE_FF
import ru.dbotthepony.kstarbound.world.TileState
import ru.dbotthepony.kvector.vector.nint.Vector2i

class RectTileFlooderSizeFirst(
	private val tiles: Array<TileState?>,
	private val seen: BooleanArray,
	private val rootx: Int,
	private val rooty: Int
) {
	val mins: Vector2i
	val maxs: Vector2i

	private fun filled(x: Int, y: Int): Boolean {
		if (x < 0 || x > CHUNK_SIZE_FF) {
			return false
		}

		if (y < 0 || y > CHUNK_SIZE_FF) {
			return false
		}

		return !seen[x or (y shl CHUNK_SHIFT)] && tiles[x or (y shl CHUNK_SHIFT)] != null
	}

	private var widthPositive = 0
	private var widthNegative = 0
	private var heightPositive = 0
	private var heightNegative = 0

	private fun checkLeft(): Boolean {
		for (i in -heightNegative .. heightPositive) {
			if (!filled(rootx - widthNegative, rooty + i)) {
				return false
			}
		}

		return true
	}

	private fun checkRight(): Boolean {
		for (i in -heightNegative .. heightPositive) {
			if (!filled(rootx + widthPositive, rooty + i)) {
				return false
			}
		}

		return true
	}

	private fun checkUp(): Boolean {
		for (i in -widthNegative .. widthPositive) {
			if (!filled(rootx + i, rooty + heightPositive)) {
				return false
			}
		}

		return true
	}

	private fun checkDown(): Boolean {
		for (i in -widthNegative .. widthPositive) {
			if (!filled(rootx + i, rooty - heightNegative)) {
				return false
			}
		}

		return true
	}

	init {
		var expanded = true
		var hitLeft = false
		var hitRight = false
		var hitUp = false
		var hitDown = false

		while (expanded) {
			expanded = false

			// expand left
			if (!hitLeft) {
				widthNegative++

				if (!checkLeft()) {
					widthNegative--
					hitLeft = true
				} else {
					expanded = true
				}
			}

			// expand up
			if (!hitUp) {
				heightPositive++

				if (!checkUp()) {
					heightPositive--
					hitUp = true
				} else {
					expanded = true
				}
			}

			// expand right
			if (!hitRight) {
				widthPositive++

				if (!checkRight()) {
					widthPositive--
					hitRight = true
				} else {
					expanded = true
				}
			}

			// expand down
			if (!hitDown) {
				heightNegative++

				if (!checkDown()) {
					heightNegative--
					hitDown = true
				} else {
					expanded = true
				}
			}
		}

		mins = Vector2i(rootx - widthNegative, rooty - heightNegative)
		maxs = Vector2i(rootx + widthPositive, rooty + heightPositive)
	}

	fun markSeen() {
		for (x in mins.x .. maxs.x) {
			for (y in mins.y .. maxs.y) {
				seen[x or (y shl CHUNK_SHIFT)] = true
			}
		}
	}
}